@@ -12,7 +12,9 @@ from .physics.obstacle import *
 from .physics.material import *
 from .physics.domain import *
 from .physics.field.effect import *
+from .physics.pressuresolver.solver_api import PoissonDomain, PoissonSolver
 from .physics.pressuresolver.sparse import SparseCG, SparseSciPy
+from .physics.pressuresolver.geom import GeometricCG
 
 from .data.fluidformat import *
 from .data.dataset import *
