@@ -1,8 +1,10 @@
+from phi import struct
 from phi.data.fluidformat import _transform_for_writing
-from .util import placeholder
 from phi.physics.physics import State
 from phi.physics.world import StateProxy
-from phi import struct
+from phi.struct.context import _unsafe
+
+from .util import placeholder
 
 
 def load_state(state):
@@ -11,7 +13,7 @@ def load_state(state):
     assert isinstance(state, State)
     state = _transform_for_writing(state)
     names = struct.names(state)
-    with struct.unsafe():
+    with _unsafe():
         placeholders = placeholder(state.shape)
     state_in = struct.map(lambda x: x, placeholders)  # validates fields, splits staggered tensors
     return state_in, {placeholders: names}
