@@ -1,12 +1,14 @@
 import logging
 import uuid
+import warnings
+
 import numpy as np
 import six
 import tensorflow as tf
 from packaging import version
 
-from phi.math.base_backend import Backend
-from phi.struct.tensorop import expand, collapsed_gather_nd
+from phi.backend.backend import Backend
+from phi.backend.tensorop import expand, collapsed_gather_nd
 
 if tf.__version__[0] == '2':
     logging.info('Adjusting for tensorflow 2.0')
@@ -19,22 +21,23 @@ class TFBackend(Backend):
     def __init__(self):
         Backend.__init__(self, "TensorFlow")
 
-    def is_applicable(self, values):
-        for value in values:
-            if self.is_tensor(value): return True
-        return False
-
     def is_tensor(self, x):
         return isinstance(x, (tf.Tensor, tf.Variable, tf.SparseTensor, tf.Operation))
 
     def as_tensor(self, x):
+        if isinstance(x, np.ndarray) and x.dtype == np.float64:
+            return tf.convert_to_tensor(x, dtype=tf.float32)
         return tf.convert_to_tensor(x)
 
     def equal(self, x, y):
         return tf.equal(x, y)
 
     def divide_no_nan(self, x, y):
-        return tf.div_no_nan(x, y)
+        if version.parse(tf.__version__) >= version.parse('1.11.0'):
+            return tf.div_no_nan(x, y)
+        else:
+            result = x / y
+            return tf.where(tf.is_finite(result), result, tf.zeros_like(result))
 
     def random_uniform(self, shape):
         return tf.random.uniform(shape)
@@ -60,7 +63,7 @@ class TFBackend(Backend):
             return self._single_mode_single_constant_pad(value, pad_width, mode, constant_values)
         else:
             mode = expand(mode, shape=(len(dims), 2))
-            passes = [('wrap', 0), ('symmetric', 0), ('reflect', 0)]
+            passes = [('circular', 0), ('wrap', 0), ('replicate', 0), ('symmetric', 0), ('reflect', 0)]
             constant_values = expand(constant_values, shape=(len(dims), 2))
             constant_value_set = set()
             for d in dims:
@@ -75,10 +78,13 @@ class TFBackend(Backend):
 
     def _single_mode_single_constant_pad(self, value, pad_width, single_mode, constant_value=0):
         single_mode = single_mode.lower()
-        assert single_mode in ('constant', 'symmetric', 'wrap', 'reflect'), single_mode
+        if single_mode == 'wrap':
+            warnings.warn("'wrap' is deprecated, use 'circular' instead", DeprecationWarning, stacklevel=2)
+            single_mode = 'circular'
+        assert single_mode in ('constant', 'symmetric', 'circular', 'reflect', 'replicate'), single_mode
         if np.sum(np.array(pad_width)) == 0:
             return value
-        if single_mode == 'wrap':
+        if single_mode == 'circular':
             dims = range(len(value.shape))
             for dim in dims:
                 s = value.shape[dim]
@@ -91,12 +97,12 @@ class TFBackend(Backend):
                 upper = value[upper_slices]
                 value = tf.concat([lower, value, upper], axis=dim)
             return value
-        else:
-            single_mode = single_mode.upper()
-            return tf.pad(value, pad_width, single_mode, constant_values=constant_value)
-
-    def add(self, values):
-        return tf.add_n(values)
+        if single_mode == 'replicate':
+            if np.any(np.array(pad_width) > 1):
+                raise NotImplementedError()  # ToDo: manual padding with slices
+            else:
+                single_mode = 'symmetric'
+        return tf.pad(value, pad_width, single_mode.upper(), constant_values=constant_value)
 
     def reshape(self, value, shape):
         return tf.reshape(value, shape)
@@ -118,11 +124,11 @@ class TFBackend(Backend):
     def where(self, condition, x=None, y=None):
         return tf.where(condition, x, y)
 
-    def mean(self, value, axis=None):
+    def mean(self, value, axis=None, keepdims=False):
         if axis is not None:
             if not isinstance(axis, int):
                 axis = list(axis)
-        return tf.reduce_mean(value, axis)
+        return tf.reduce_mean(value, axis, keepdims=keepdims)
 
     def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
         if grad is None:
@@ -139,8 +145,12 @@ class TFBackend(Backend):
             result.set_shape(shape_out)
         return result
 
-    def resample(self, inputs, sample_coords, interpolation="LINEAR", boundary="zero"):
-        return resample_tf(inputs, sample_coords, interpolation, boundary)
+    def resample(self, inputs, sample_coords, interpolation='linear', boundary='constant'):
+        if boundary.lower() == 'constant':
+            boundary = 'zero'
+        boundary_func = SUPPORTED_BOUNDARY[boundary.lower()]
+        assert interpolation.lower() == 'linear'
+        return _resample_linear_niftynet(inputs, sample_coords, boundary, boundary_func)
 
     def zeros_like(self, tensor):
         return tf.zeros_like(tensor)
@@ -257,6 +267,8 @@ class TFBackend(Backend):
         return tf.to_complex64(x)
 
     def gather(self, values, indices):
+        if isinstance(indices, slice):
+            return values[indices]
         return tf.gather(values, indices)
 
     def gather_nd(self, values, indices):
@@ -310,6 +322,7 @@ class TFBackend(Backend):
     def fft(self, x):
         rank = len(x.shape) - 2
         assert rank >= 1
+        x = self.to_complex(x)
         if rank == 1:
             return tf.stack([tf.fft(c) for c in tf.unstack(x, axis=-1)], axis=-1)
         elif rank == 2:
@@ -349,6 +362,9 @@ class TFBackend(Backend):
     def dtype(self, array):
         return array.dtype.as_numpy_dtype
 
+    def sparse_tensor(self, indices, values, shape):
+        return tf.SparseTensor(indices=indices, values=values, dense_shape=shape)
+
 
 # from niftynet.layer.resampler.py
 # https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyNet/blob/69c98e5a95cc6788ad9fb8c5e27dc24d1acec634/niftynet/layer/resampler.py
@@ -427,20 +443,6 @@ def _resample_linear_niftynet(inputs, sample_coords, boundary, boundary_func):
     return _pyramid_combination(samples, weight_0, weight_1)
 
 
-def resample_tf(inputs, sample_coords, interpolation="LINEAR", boundary="zero"):
-    """
-Resamples an N-dimensional tensor at the locations provided by sample_coords
-    :param inputs: grid with dimensions (batch_size, spatial dimensions..., element_size)
-    :param sample_coords: sample coords (batch_size, output_shape, input_dimension)
-    :param interpolation: LINEAR, BSPLINE, IDW (default is LINEAR)
-    :param boundary: ZERO, REPLICATE, CIRCULAR, SYMMETRIC (default is ZERO)
-    :return:
-    """
-    boundary_func = SUPPORTED_BOUNDARY[boundary.lower()]
-    assert interpolation.upper() == "LINEAR"
-    return _resample_linear_niftynet(inputs, sample_coords, boundary, boundary_func)
-
-
 def _boundary_snap(sample_coords, spatial_shape):
     max_indices = [l-1 for l in spatial_shape]
     for _i in range(len(spatial_shape)):
@@ -459,14 +461,19 @@ def _boundary_circular(sample_coords, input_size):
 
 
 def _boundary_symmetric(sample_coords, input_size):
-    circular_size = input_size + input_size - 2
-    return (input_size - 1) - tf.abs(
-        (input_size - 1) - _boundary_circular(sample_coords, circular_size))
+    sample_coords = _boundary_circular(sample_coords, 2 * input_size)
+    return ((2 * input_size - 1) - tf.abs((2 * input_size - 1) - 2 * sample_coords)) // 2
+
+
+def _boundary_reflect(sample_coords, input_size):
+    sample_coords = _boundary_circular(sample_coords, 2 * input_size - 2)
+    return (input_size - 1) - tf.abs((input_size - 1) - sample_coords)
 
 
 SUPPORTED_BOUNDARY = {
     'zero': _boundary_replicate,
     'replicate': _boundary_replicate,
     'circular': _boundary_circular,
-    'symmetric': _boundary_symmetric
+    'symmetric': _boundary_symmetric,
+    'reflect': _boundary_reflect,
 }
